/* regd.cc
 * Written by Joshua Rowe
 */

#include	"key.hh"
#include	"reglib.h"
#include	<errno.h>
#include	<fpios.hh>
#include	<getopt.h>
#include	<netinet/in.h>
#ifdef	SHADOW_PWD
#include	<shadow.h>
#endif
#include	<pwd.h>
#include	<signal.h>
#include	<stdarg.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/socket.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<sys/utsname.h>
#include	<sys/wait.h>
#include	<unistd.h>

#define	OPT_NO_DETACH	1000
#define	OPT_DEBUG	1001

struct	statestruct
{
  int		state;		// enumstate value
  int		bytesleft;	// bytes left in current read
  void *	byteptr;	// next place to put data
  int		request;	// request to make
  int		pathlen;	// length of path
  char *	path;		// path to operate on
  int		type;		// type of data
  int		datalen;	// length of data
  void *	data;		// data to use
  int		uid;		// user id of the connection
  int		gid;		// gid of the connection
  RegKey *	userskey;	// user's personal key
  char *	pwnam;		// user's name
};

enum	enumstate {
  state_readreq,
  state_readpathlen,
  state_readpath,
  state_readtype,
  state_readdatalen,
  state_readdata,
  state_end
  };

char * *	args			= 0;
RegKey *	defaultu		= NULL;
int		debugging		= 0;
int		detachflag		= 2;
int		numberofconnections	= 1;
int		port			= 5500;
pid_t		sid			= -1;
int		sock			= -1;
fd_set		readhandles;
int		readsema4		= 0;
char *		regfn			= NULL;
pRegKey		registry		= NULL;
int		sigtermsema4		= 0;
statestruct	states[FD_SETSIZE]; // these do not need to be initialized
pRegKey		systemu			= NULL;
pRegKey		users			= NULL;
int		watchdog		= 300;
int		watchdogp		= 0;
int		writesema4		= 0;
int		wtransaction		= 1;

void	acceptconnection();
void	closesocket();
void	createkey(int i);
void	debug(const char *c, ...);
void	derror(const char *c);
void	detach();
void	existkey(int i);
void	existval(int i);
void	getfirstkey(int i);
void	getfirstval(int i);
RegKey *getkey(int i, const char *path);
void	getnextkey(int i);
void	getnextval(int i);
RegVal *getval(int i, const char *path);
void	getvalue(int i);
void	killkey(int i);
void	killval(int i);
void	mainloop();
int	match(const char *path, const char *name);
void	opensocket();
void	printusage(char *arg0);
void	printversion();
void	processrequest(int i);
int	readdata(int i);
void	readregistry();
void	setkeyinfo(int i);
void	setuphandlers();
void	setupsysreg();
void	setupwatchdog();
void	setval(int i);
void	sighup_handler(int sig);
void	sigpipe_handler(int sig);
void	sigterm_handler(int sig);
void	sigusr1_handler(int sig);
void	writeregistry();

void *	operator new(size_t t)
{
  return malloc(t);
}

void	operator delete(void *x, size_t t)
{
  free(x);
}

void	acceptconnection()
{
  debug("acceptconnection\n");
  int	nw	= accept(sock, NULL, 0);
  if (nw < 0)
    {
      derror("accept");
      return;
    }
  FD_SET(nw, &readhandles);
  states[nw].state	= state_readreq;
  states[nw].bytesleft	= sizeof(states[nw].request);
  states[nw].byteptr	= &states[nw].request;
  states[nw].uid	= -1;
  states[nw].gid	= -1;
  states[nw].userskey	= defaultu;
  states[nw].pwnam	= NULL;
}

void	closesocket()
{
  debug("closesocket\n");
  close(sock);
}

void	createkey(int i)
{
  debug("createkey");
  char *	pe	= strrchr(states[i].path, '/');
  if (pe == NULL)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  *pe	= 0;
  RegKey *	d	= getkey(i, states[i].path);
  if (d == NULL)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  *pe	= '/';
  if (!d->access(states[i].uid, states[i].gid, RKM_W))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  if (getkey(i, states[i].path))
    {
      regd_writeint(i, RG_ERR_EXIST);
      return;
    }
  new RegKey(pe, states[i].uid, states[i].gid, states[i].type, d);
  wtransaction++;
  regd_writeint(i, RG_ERR_SUCCESS);
}

void	debug(const char *c, ...)
{
  if (!debugging)
    return;
  va_list	x;
  fprintf(stderr, "%s: ", args[0]);
  va_start(x, c);
  vfprintf(stderr, c, x);
  va_end(x);
  fprintf(stderr, "\n");
}

void	derror(const char *c)
{
  if (debugging)
    perror(c);
}

void	detach()
{
  debug("detach");
  int	x;
  chdir("/");
  if ((x = fork()) > 0)
    {
      sleep(1);			// Make sure child can detach (1 second
				// or so should do it...
      exit(EXIT_SUCCESS);	// Parent
    }
  else if (x == -1)
    {				// Unsuccessful
      debug("regd:  Unable to fork new process!");
      derror("fork");
      exit(1);
    }
  if ((sid = setsid()) == -1)
    {				// Successful, now remove ourselves from
				// the real world
      debug("regd:  setsid failed!");
      derror("setsid");
      exit(1);
    }
}

void	existkey(int i)
{
  RegKey *	d;
  if (!(d = getkey(i, states[i].path)))
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (!d->parent)
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      return;
    }
  else if (!d->parent->access(states[i].uid, states[i].gid, RKM_R))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  else
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      return;
    }
}

void	existval(int i)
{
  RegVal *	d;
  if (!(d = getval(i, states[i].path)))
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (!d->parent)
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      return;
    }
  else if (!d->parent->access(states[i].uid, states[i].gid, RKM_R))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  else
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      return;
    }
}

void	getfirstkey(int i)
{
  RegKey *	d	 = getkey(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (d->children == 0)
    regd_writeint(i, RG_ERR_NOMORE);
  else if (d->access(states[i].uid, states[i].gid, RKM_R))
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      char *	name	= d->children->getfullname();
      regd_writestring(i, name);
      delete	name;
    }
  else
    regd_writeint(i, RG_ERR_NOACC);
}

void	getfirstval(int i)
{
  RegKey *	d	 = getkey(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (d->vals == 0)
    regd_writeint(i, RG_ERR_NOMORE);
  else if (d->access(states[i].uid, states[i].gid, RKM_R))
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      char *	name	= d->vals->getfullname();
      regd_writestring(i, name);
      delete	name;
    }
  else
    regd_writeint(i, RG_ERR_NOACC);
}

RegKey *	getkey(int i, const char *path)
{
  RegKey *	curkey;
  RegKey *	thekey;
  if (strlen(path) == 0)
    return registry;
  if (match(path, "/me"))	// the path "/me ..." special case
    {
      curkey	= states[i].userskey->children;
      path	+= strlen("/me");
      thekey	= states[i].userskey;
    }
  else
    {
      curkey	= registry->children;
      thekey	= registry;
    }
  while ((path[0] != 0) && (path[0] != '|') && (curkey != NULL))
    if (match(path, curkey->name))
      {
	thekey	= curkey;
	path		+= strlen(thekey->name);
	curkey	= thekey->children;
      }
    else
      curkey	= curkey->next;
  if ((path[0] != 0) && (path[0] != '|'))
    return	NULL;
  return	thekey;
}

void	getkeyinfo(int i)
{
  debug("getkeyinfo");
  RegKey *	d	= getkey(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if ((d->parent) && (!d->parent->access(states[i].uid,
					 states[i].gid,
					 RKM_R)))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  regd_writeint(i, RG_ERR_SUCCESS);
  char *	c	= d->getfullname();
  regd_writestring(i, c);
  delete	c;
  regd_writeint(i, d->uid);
  regd_writeint(i, d->gid);
  regd_writeint(i, d->mode);
}

void	getnextkey(int i)
{
  RegKey *	d	= getkey(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (d->next == 0)
    regd_writeint(i, RG_ERR_NOMORE);
  else if ((d->parent) && (d->parent->access(states[i].uid,
					     states[i].gid,
					     RKM_R)))
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      char *	name	= d->next->getfullname();
      regd_writestring(i, name);
      delete	name;
    }
  else
    regd_writeint(i, RG_ERR_NOACC);
}

void	getnextval(int i)
{
  RegVal *	d	= getval(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (d->next == 0)
    regd_writeint(i, RG_ERR_NOMORE);
  else if ((d->parent) && (d->parent->access(states[i].uid,
					     states[i].gid,
					     RKM_R)))
    {
      regd_writeint(i, RG_ERR_SUCCESS);
      char *	name	= d->next->getfullname();
      regd_writestring(i, name);
      delete	name;
    }
  else
    regd_writeint(i, RG_ERR_NOACC);
}

RegVal *getval(int i, const char *path)
{
  RegKey *	p	= getkey(i, path);
  if (p == 0)
    return NULL;
  path	= strrchr(path, '|');
  if (path == NULL)
    return NULL;
  RegVal *	q	= p->vals;
  while (q && strcmp(q->name, path))
    q	= q->next;
  return q;
}

void	getvalue(int i)
{
  RegVal *	p	= getval(i, states[i].path);
  if (p == NULL)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  regd_writeint(i, RG_ERR_SUCCESS);
  regd_writestring(i, p->data);
}

void	killkey(int i)
{
  RegKey *	d	= getkey(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (d->parent)
    if (d->parent->mode & RKM_STICKY)
      {
	if (!d->access(states[i].uid, states[i].gid, RKM_W))
	  {
	    regd_writeint(i, RG_ERR_NOACC);
	    return;
	  }
      }
    else
      if (!d->parent->access(states[i].uid, states[i].gid, RKM_W))
	{
	  regd_writeint(i, RG_ERR_NOACC);
	  return;
	}
  delete	d;
  regd_writeint(i, RG_ERR_SUCCESS);
}

void	killval(int i)
{
  RegVal *	d	= getval(i, states[i].path);
  if (d == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if ((d->parent) &&
      (!d->parent->access(states[i].uid, states[i].gid, RKM_W)))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  delete	d;
  regd_writeint(i, RG_ERR_SUCCESS);
}

void	login(int i)
{
  struct passwd *p	= getpwnam(states[i].path);
  if (p == NULL)
    {
      regd_writeint(i, RG_ERR_BADPASS);
      return;
    }  
  char *	c	= p->pw_passwd;
#ifdef	SHADOW_PWD
  struct spwd	*sp;
  sp		= getspnam(states[i].path);
  if (sp == NULL)
    {
      regd_writeint(i, RG_ERR_BADPASS);
      return;
    }
  c		= sp->sp_pwdp;
#endif
  if (strcmp(crypt((char *)states[i].data, c), c))
    {
      regd_writeint(i, RG_ERR_BADPASS);
      return;
    }
  if (states[i].pwnam)
    free(states[i].pwnam);
  states[i].uid		= p->pw_uid; // All that just for this?!!?! heh heh heh
  states[i].gid		= p->pw_gid;
  states[i].pwnam	= strdup(p->pw_name);
  char	home[50];
  sprintf(home, "/users/%i", states[i].uid);
  if ((states[i].userskey = getkey(i, home)) == 0)
    {
				// This is the one place that the user
				// has access permissions
      sprintf(home, "/%i", states[i].uid);
      states[i].userskey	= new RegKey(home,
					     states[i].uid,
					     states[i].gid,
					     RKM_DEF,
					     users);
      wtransaction++;
    }
  regd_writeint(i, RG_ERR_SUCCESS);
}

void	mainloop()		// This is where all the real work gets done
{
  debug("mainloop");
  while (1)
    {
      if (writesema4)
	writeregistry();
      if (readsema4)
	{
	  delete registry;
	  readregistry();
	}
      if (sigtermsema4)
	{
	  sigtermsema4	= 2;
	  raise(SIGTERM);
	}
				// Make a copy of the file handle info
      fd_set	readhandles_sel	= readhandles;
      int	exitflag	= 0;
      int	retval		= select(FD_SETSIZE, &readhandles_sel,
					 NULL, NULL, NULL);
      if (retval < 0)
	{
	  derror("select");
	  switch (errno)
	    {
	    case EINTR:
	      if (writesema4)
		writeregistry();
	      if (readsema4)
		{
		  delete registry;
		  readregistry();
		}
	      if (sigtermsema4)
		{
		  sigtermsema4	= 2;
		  raise(SIGTERM);
		}
	      break;
	    default:
	      exitflag++;
	      break;
	    }
	}
      else
	for (int i=0; i < FD_SETSIZE; i++)
	  if (FD_ISSET(i, &readhandles_sel))
	    if (i == sock)
	      acceptconnection();
	    else
	      if (readdata(i)) // readdata returns 1 if request is
		processrequest(i); // now pending, so go process it
      if (exitflag)
	break;
      retval	= 0;
    }  
}

// Returns true if name is the first component of path
int	match(const char *path, const char *name)
{
  char *	c	= strstr(path, name);
  int	x		= strlen(name);
  if (c != path)
    return 0;
  if ((path[x] == 0) || (path[x] == '/') || (path[x] == '|'))
    return 1;
  else
    return 0;
}

void	opensocket()
{
  debug("opensocket");
				// The following code is copied from the
				// GNU glibc info manual... thanks guys!
  struct sockaddr_in name;
  
  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    {
      derror ("socket");
      exit (EXIT_FAILURE);
    }
  
  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons (port);
  name.sin_addr.s_addr = htonl (INADDR_ANY);
  if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
    {
      derror ("bind");
      exit (EXIT_FAILURE);
    }
				// Code from here is mine
  if (listen(sock, numberofconnections) < 0)
    {
      derror("listen");
      exit (EXIT_FAILURE);
    }
  FD_ZERO(&readhandles);	// Zero out the fd_set
  FD_SET(sock, &readhandles);	// Set the socket handle

  states[sock].userskey	= registry;
  states[sock].uid	= 0;
  states[sock].gid	= 0;
}

void	printusage(char *arg0)
{
  fprintf(stderr,
	  "usage - %s [OPTION]...\n"
	  "\n"
	  "  -d, --detach N        detach from terminal if N is 1\n"
	  "  -f, --file F          use file F\n"
	  "  -h, --help            output this screen and exit\n"
	  "  -n, --number-of-conns N number of connections in queue (1)\n"
	  "  -p, --port P          listen on port P\n"
	  "  -v, --version         output version information and exit\n"
	  "  -w, --watchdog W      cause a write every W seconds\n"
	  "      --no-detach       don't detach from terminal\n"
	  "      --detach          detach from terminal (default)\n"
	  "\n"
	  "Without a --file, " REGDB " is used instead.\n"
	  ,arg0
	  );
}

void	printversion()
{
  printf("regd version " VERNUM "\nWritten by Joshua Rowe.\n");
}

void	processrequest(int i)
{
  debug("processrequest");
  switch(states[i].request)
    {
    case RGQ_FIRSTKEY:
      getfirstkey(i);
      break;
    case RGQ_NEXTKEY:
      getnextkey(i);
      break;
    case RGQ_LOGIN:
      login(i);
      break;
    case RGQ_CREATEKEY:
      createkey(i);
      break;
    case RGQ_EXISTKEY:
      existkey(i);
      break;
    case RGQ_GETKEYI:
      getkeyinfo(i);
      break;
    case RGQ_KILLKEY:
      killkey(i);
      break;
    case RGQ_FIRSTVAL:
      getfirstval(i);
      break;
    case RGQ_NEXTVAL:
      getnextval(i);
      break;
    case RGQ_SETKEYI:
      setkeyinfo(i);
      break;
    case RGQ_KILLVAL:
      killval(i);
      break;
    case RGQ_SETVAL:
      setval(i);
      break;
    case RGQ_GETVAL:
      getvalue(i);
      break;
    case RGQ_EXISTVAL:
      existval(i);
      break;
    default:
      regd_writeint(i, RG_ERR_INVREQ);
    }
  delete		states[i].path;	// Reset to a cool state
  delete		states[i].data;
  states[i].state	= state_readreq;
  states[i].bytesleft	= sizeof(states[i].request);
  states[i].byteptr	= &states[i].request;
}

int	readdata(int i)
{
  debug("readdata");
  int	x		= read(i, states[i].byteptr, states[i].bytesleft);
  if (x == 0)			// EOF?
    {
      debug("closeconn %i", i);
      close(i);
      FD_CLR(i, &readhandles);
      if (states[i].pwnam != 0)
	free(states[i].pwnam);
      return 0;			// We'll never here from him again...
    }
  states[i].bytesleft	-= x;
  states[i].byteptr	+= x;
  if (states[i].bytesleft == 0)
    {
      states[i].state++;
      switch(states[i].state)
	{
	case state_readpathlen:
	  states[i].bytesleft		= sizeof(states[i].pathlen);
	  states[i].byteptr		= &(states[i].pathlen);
	  break;
	case state_readpath:
	  states[i].path		= new char[states[i].pathlen + 1];
	  states[i].path[states[i].pathlen]	= 0; // no GPF's later on...
	  states[i].bytesleft		= states[i].pathlen;
	  states[i].byteptr		= states[i].path;
	  break;
	case state_readtype:
	  states[i].bytesleft		= sizeof(states[i].type);
	  states[i].byteptr		= &(states[i].type);
	  break;
	case state_readdatalen:
	  states[i].bytesleft		= sizeof(states[i].datalen);
	  states[i].byteptr		= &(states[i].datalen);
	  break;
	case state_readdata:
	  states[i].data		= new char[states[i].datalen + 1];
	  ((char *)(states[i].data))[states[i].datalen]	= 0;
				// That previous line is for the functions
				// which interpret data as a string
	  states[i].bytesleft		= states[i].datalen;
	  states[i].byteptr		= states[i].data;
	  break;
	case state_end:
	  return 1;		// Ready for request now
	default:
	  debug("error in checking states...");
	  exit(EXIT_FAILURE);	// This should never happen
	}
    }
  return 0;
}

void	readregistry()
{
  char	buf[32768];
  debug("readregistry");
  ifpstream	is(regfn);
  is >> registry >> users >> defaultu >> systemu;
  if ((!users) || (!registry) || (!is))
    {
      derror("readregistry");
      exit(EXIT_FAILURE);			// Errors reading registry file
    }
  readsema4	= 0;
}

void	setkeyinfo(int i)
{
  RegKey *p	= getkey(i, states[i].path);
  if (p == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if ((states[i].uid == 0) || (states[i].uid == p->uid))
    p->mode	= states[i].type;
  else
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  if (states[i].datalen == (sizeof(int) * 2))
    if (states[i].uid == 0)
      {
	regd_writeint(i, RG_ERR_SUCCESS);
	p->uid	= ((int *)states[i].data)[0];
	p->gid	= ((int *)states[i].data)[1];
      }
    else
      {
	regd_writeint(i, RG_ERR_NOACC);
	return;
      }
  else
    if (states[i].datalen == 0)
      {
	regd_writeint(i, RG_ERR_SUCCESS);
	return;
      }
    else
      {
	regd_writeint(i, RG_ERR_INVREQ);
	return;
      }
}

void	setuphandlers()
{
  debug("setuphandlers");
  signal(SIGHUP, sighup_handler);
  signal(SIGPIPE, sigpipe_handler);
  signal(SIGTERM, sigterm_handler);
  signal(SIGUSR1, sigusr1_handler);
  signal(SIGINT, sigterm_handler);
}

void	setupsysreg()
{
  if (systemu)
    delete	systemu;
  systemu	= new RegKey("/system", 0, 0, RKM_DEF, registry);
  struct utsname	x;
  uname(&x);
  pRegKey	osk	= new RegKey("/os", 0, 0, RKM_DEF, systemu);
  new RegVal("|sysname", x.sysname, osk);
  new RegVal("|nodename", x.nodename, osk);
  new RegVal("|release", x.release, osk);
  new RegVal("|version", x.version, osk);
  new RegVal("|machine", x.machine, osk);
}

void	setupwatchdog()
{
  int	watchdogp	= fork();
  debug("setupwatchdog");
  if (watchdogp == 0)
    {
      closesocket();
      int	x	= getppid();
      while (1)
	{
	  debug("watchdog");
	  sleep(watchdog);
	  kill(x, SIGUSR1);
	}
    }
  else if (watchdogp < 0)
    {
      fprintf(stderr, "Unable to set up watchdog... proceding w/o it!\n");
      derror("fork");
    }
}

void	setval(int i)
{
  RegVal *	p	= getval(i, states[i].path);
  RegKey *	k	= getkey(i, states[i].path);
  if (k == 0)
    {
      regd_writeint(i, RG_ERR_NONAME);
      return;
    }
  if (!k->access(states[i].uid, states[i].gid, RKM_W))
    {
      regd_writeint(i, RG_ERR_NOACC);
      return;
    }
  if (p == 0)
    new RegVal(strchr(states[i].path, '|'), (char *)(states[i].data), k);
  else
    {
      if (p->data)
	delete p->data;
      p->data	= new char[strlen((char *)(states[i].data)) + 1];
      strcpy(p->data, (char *)(states[i].data));
    }
  regd_writeint(i, RG_ERR_SUCCESS);
  return;
}

// A SIGHUP will cause the regd program to write the registry, then
// to re-read it.  Rereading causes the regd to reconstruct volatile
// structures.  
void	sighup_handler(int sig)
{
  writesema4++;
  readsema4++;
  setuphandlers();
}

// Just ignore a sigpipe for now.  This should only happen if we are in
// the middle of a transaction and we haven't detected the EOF yet.  
void	sigpipe_handler(int sig)
{
  setuphandlers();
}

void	sigterm_handler(int sig)
{
  if (sigtermsema4 == 2)
    {
      kill(watchdogp, sig);
      raise(sig);
    }
  else
    {
      sigtermsema4	= 1;
      writesema4++;
      setuphandlers();
    }
}

void	sigusr1_handler(int sig)
{
  writesema4++;
  setuphandlers();
}

void	sigusr2_handler(int sig)
{
  signal(SIGUSR2, sigusr2_handler);
}

void	writeregistry()
{
  debug("writeregistry");
  if (!wtransaction)
    return;			// No write transactions processed and not
				// yet written
  char *	newfn	= new char[strlen(regfn) + 2];
  newfn[strlen(regfn) + 1]	= 0;
  strcpy(newfn, regfn);
  newfn[strlen(regfn)]	= (regfn[strlen(regfn)] == '~') ? '_' : '~';
  ofpstream	os(newfn, pstream::xxtrunc | pstream::xxcreat);
  os << registry << users << defaultu << systemu;
  if (!os)
    {
      derror("writeregistry");
      exit(1);			// Couldn't write registry.
    }
				// Uh oh...
  os.close();			// Flushes the file.
  rename(newfn, regfn);		// Atomically recreates the file.
  delete	newfn;
  wtransaction	= 0;		// Tell future writeregistry's not to write
}

int	main(int argc, char **argv)
{
  args	= argv;
  regfn	= new char[strlen(REGDB) + 1];
  strcpy(regfn, REGDB);
  int	c;
  while (1)
    {
      static struct option long_options[] =
	{
	  {"debug",	no_argument,		NULL,	OPT_DEBUG},
	  {"detach",	optional_argument,	NULL,  	'd'},
	  {"file",	required_argument,	NULL,	'f'},
	  {"help",	no_argument,		NULL,	'h'},
	  {"no-detach",	no_argument,		NULL,  	OPT_NO_DETACH},
	  {"number-of-conns", required_argument,NULL,	'n'},
	  {"port",	required_argument,	NULL,	'p'},
	  {"version",	no_argument,		NULL,	'v'},
	  {"watchdog",	required_argument,	NULL,	'w'},
	  {0,		0,			NULL,	0}
	};
      c	= getopt_long(argc, argv, "d:f:hn:p:vw:", long_options, NULL);
      if (c == -1)
	break;			// We have reached the end of the options
      switch(c)
	{
	case 'd':		// -d or --detach
	  if (optarg)
	    {
	      if (strcmp(optarg, "1") == 0)
		detachflag	= 1; // user told us to detach
	      else if (strcmp(optarg, "0") == 0)
		detachflag	= 0; // user told us not to detach
	      else
		{		// user didn't know what he was saying
		  fprintf(stderr,
			  "-d, --detach can take only 0 or 1 as an argument!\n"
			  );
		  return 1;
		}
	    }
	  else
	    detachflag	= 1;	// default is to detach from terminal
	  break;
	case 'f':		// -f or --file
	  delete	regfn;
	  regfn		= new char[strlen(optarg) + 1];
	  strcpy(regfn, optarg);
	  break;
	case 'h':		// -h or --help
	  printusage(argv[0]);
	  return EXIT_SUCCESS;
	case 'n':		// -n or --number-of-conns
	  numberofconnections	= strtol(optarg, NULL, 0);
	  if (numberofconnections < 1)
	    {
	      fprintf(stderr, "-n, --number-of-conns must be a number greater than 0\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case 'p':		// -p or --port
	  port	= strtol(optarg, NULL, 0);
	  if (port < 0)
	    {
	      fprintf(stderr, "-p, --port must be a number greater than 0\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case 'v':		// --version
	  printversion();
	  return EXIT_SUCCESS;
	case 'w':
	  watchdog	= strtol(optarg, NULL, 0);
	  if (watchdog < 0)
	    {
	      fprintf(stderr, "-w, --watchdog must be a number greater than 0\n");
	      exit(EXIT_FAILURE);
	    }
	  break;
	case OPT_NO_DETACH:	// --no-detach
	  detachflag	= 0;
	  break;
	case OPT_DEBUG:
	  debugging	= 1;
	  if (detachflag == 2)
	    detachflag	= 0;
	  break;
	case '?':		// invalid option
	  printusage(argv[0]);
	  return EXIT_FAILURE;
	}
    }
  if (regd_connect(0, port) > 0)
    {
      fprintf(stderr, "Someone is already listening to port %i!\n", port);
      exit(1);
    }
  if (detachflag)
    detach();
  wtransaction	= 0;		// Make sure that a database doesn't
				// get written until after we've read a
				// valid image
  opensocket();
  setupwatchdog();
  setuphandlers();
  readregistry();
  mainloop();
  kill(watchdogp, SIGTERM);
  waitpid(watchdogp, NULL, 0);
  closesocket();
  writeregistry();
}
